#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include "my_index.h"
#include "dict.h"

void main(int argc, char **argv);

void main(int argc, char **argv)
{
dictnode *dictionary;
char phrase[80];  /* overly large -- i hope card names are never this long! */
char **names;     /* an array of strings to hold the card names */
FILE *phraselist;
int errorcount;   /* number of errors returned on a match from alookup() */
int id;           /* an id number to tag all card names with */
int numnames = 0;

if (argc != 2)
  {
    fprintf(stderr,"usage: %s phrase_list\n",argv[0]);
    exit(1);
  }

phraselist = fopen(argv[1],"r");

if(phraselist == NULL)
  {
    fprintf(stderr,"error: cannot open %s for reading\n",argv[1]);
    exit(1);
  }

/* first, read in the strings and make a dictionary of card names. */
dictionary = init_dict(ALPHABET_SIZE, my_index);  /* get a new dictionary */
names = (char **)malloc(1000 * sizeof(char *));  /* max # of entries = 1000 */

memset(phrase,'\0',80);  /* clear out phrase */
while(fgets(phrase,80,phraselist) != NULL)  /* for all phrases in the file */
  {
    char *c;

    phrase[strlen(phrase)-1] = '\0'; /* smash the newline at end of phrase */
    for(c = phrase; *c; c++) *c = tolower(*c); /* lowercase the string */

    names[numnames]=(char *)malloc((strlen(phrase)+1) * sizeof(char));
    strcpy(names[numnames],phrase);

#ifdef DEBUG
    fprintf(stderr,"names[i]: %s, phrase: %s\n",names[numnames],phrase);
#endif

    if (insert_phrase(dictionary,phrase,numnames) == -1) /* insert it */
      fprintf(stderr,"error inserting %s into dictionary. skipping.\n",phrase);
    numnames++;
    memset(phrase,'\0',80);  /* clear out phrase for next time */
  }

#ifdef DEBUG
fprintf(stderr,"finished building dictionary.  %d words\n",numnames);
#endif

/* get some user input phrases and test the dictionary */
while(fgets(phrase,80,stdin) != NULL)
  {
    char *c;
    errorcount = 0;
    phrase[strlen(phrase)-1] = '\0';
    for(c = phrase; *c; c++)
      {
	*c = tolower(*c); /* lowercase the string */
	if(*c != ' ' && (*c < 'a' && *c > 'z'))    /* bad character in input */
	  {
	    fprintf(stderr,"bad input!  ignored.\n");
	    break;
	  }
      }
    if((id = alookup(dictionary,phrase,&errorcount)) == -1)
      printf("%s was not found\n",phrase);
    else
      printf("%s mapped to %s with %d error%s\n",phrase,names[id],errorcount,
	     (errorcount == 1)?".":"s.");
  }

dealloc_dict(dictionary);
exit(0);
}
