#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include "my_index.h"
#include "dict.h"

typedef struct pricenode {
  float price;
  struct pricenode *next;
} pricelist;

int main(int argc, char **argv);
void cleanup(char *line);
char **separate_line(char *line,int *num_words);
float find_price(char **words, int num_words);

int main(int argc, char **argv)
{
dictnode *dictionary;
char phrase[80];  /* overly large -- i hope card names are never this long! */
char line[256];   /* buffer to hold citations */
char **names;     /* an array of strings to hold the card names */
FILE *phraselist, *datafile, *lastweek;
int errorcount;   /* number of errors returned on a match from alookup() */
int id;           /* an id number to tag all card names with */
int numnames = 0; /* how many names in the phraselist */
float *rawtotal;  /* array of price totals, unfiltered */
float *squaretotal;  /* array of price squared totals, unfiltered */
int *hits;        /* array of how many prices for each id */
pricelist **prices;/* array of linked lists of prices so we can do staistics */
pricelist **pends;/* pointers to the ends of the prices lists for efficiency */
int done = 0;
float *lastprices;/* prices from the lastweek file */
int i = 0;        /* everybody's favorite counter */
int dbg_id = -1;  /* if this isn't -1, enable debugging for that id */
int status = 0;   /* used to let user know we're still running */
int tindex = -1,  /* used to identify multiple names that are the same card */
    index = -1;
float tfiltavg = 0.0, /* totals to be printed at the bottom of the list */
      tstddev  = 0.0,
      trawavg  = 0.0,
      thigh    = 0.0,
      tlow     = 0.0,
      tchange  = 0.0;
int   thits    = 0;
int shadis = strcmp(argv[0],"pricelist"); /* for output formatting later */
int debug = !strcmp(argv[0],"debuglist"); /* for verbose debugging       */
int have_old_data = 0;    /* flag for whether to calculate change column */

if (argc < 3)
  {
    fprintf(stderr,"usage: %s phrase_list data_file [last_week [debug card]] > output_file\n",argv[0]);
    exit(1);
  }

phraselist = fopen(argv[1],"r");
if(phraselist == NULL)
  {
    fprintf(stderr,"error: cannot open %s for reading\n",argv[1]);
    exit(1);
  }

datafile = fopen(argv[2],"r");
if (datafile == NULL)
  {
    fprintf(stderr,"error: cannot open %s for reading\n",argv[2]);
    exit(1);
  }

if(argc >= 4)
  lastweek = fopen(argv[3],"r");
else
  lastweek = NULL;
if (lastweek == NULL)
  {
    fprintf(stderr,"warning: cannot open %s for reading\n",argv[3]);
    fprintf(stderr,"change column will contain 0.00 values\n");
  }
else
  have_old_data = 1;

if(debug)
  { fprintf(stderr,"opened data files\n"); fflush(stderr); }

/* first, read in the strings and make a dictionary of card names. */
dictionary = init_dict(ALPHABET_SIZE, my_index);  /* get a new dictionary */
fscanf(phraselist,"%d\n",&numnames); /* how many distinct names in the file */
names = (char **)malloc(numnames * sizeof(char *));

memset(phrase,'\0',80);  /* clear out phrase */
#ifdef DEBUG
fprintf(stderr,"reading phrases\n");
#endif
/* for all phrases in the file... */
while(!done)
  {
    char tag;
    int j;

    if((tag = fgetc(phraselist)) == EOF) /* read the first char on the line */
      {                                  /* and test for end of file */
	done = 1;
	continue;
      }

    switch(tag)  /* maintain index and tindex in accorance with the tags */
      {          /* on each card name */
      case '-':
	tindex = -2;
	break;
      case '+':
	tindex = ++index;
	break;
      case ' ':
	if(tindex == -2) /* a - phrase not followed by a + phrase. */
	  {
	    fprintf(stderr,"badly formed phrase list at %s\n",phrase);
	    exit(1);
	  }
	break;
      default:
	fprintf(stderr,"bad tag: |%c|\n",tag);
	exit(1);
      }

    j = 0;
    while((phrase[j++] = fgetc(phraselist)) != '\n'); /* read a phrase */
    phrase[strlen(phrase)-1] = '\0'; /* smash the newline at end of phrase */

    if(tindex != -1) /* things to ignore have this index, so don't store */
                     /* them in names[] */
      if(names[tindex] == NULL)  /* if this is a new index */
	{
	  names[tindex]=(char *)malloc((strlen(phrase)+1) * sizeof(char));
	  strcpy(names[tindex],phrase);
	}

    if (insert_phrase(dictionary,phrase,tindex) == -1) /* insert it */
      fprintf(stderr,"error inserting %s into dictionary. skipping.\n",phrase);
    memset(phrase,'\0',80);  /* clear out phrase for next time */
  }

if(debug)
  { fprintf(stderr,"read phrase list\n"); fflush(stderr); }

/* read and remember last week's prices */
lastprices = (float *)malloc(numnames * sizeof(float));
for(i=0; i < numnames; i++)
  lastprices[i] = -1.0; /* sentinal value for items not found */

if(have_old_data)
  {
#ifdef DEBUG
    fprintf(stderr,"reading last week's prices\n");
#endif
    fgets(line,256,lastweek); /* junk the top three lines */
    fgets(line,256,lastweek);
    fgets(line,256,lastweek);
#ifdef DEBUG2
    fprintf(stderr,"junked top two lines\n");
#endif
    while(fgets(line,256,lastweek) != NULL) 
      {
	char tempname[80];  /* excessively large */
	int j;
	
	j = 0;
	
#ifdef DEBUG2
	fprintf(stderr,"line: %s",line);
#endif

	while((tempname[j] = line[j]) != ' ' || line[j-1] != ' ') j++;
	tempname[j-2] = '\0'; /* smash the comma on the end of the line */
	
#ifdef DEBUG2
	fprintf(stderr,"tempname: %s",tempname);
#endif
	
	if(tempname[0] == 't' && !strcmp(tempname,"total"))/* a "total" hack */
	  continue;
	
	tindex = alookup(dictionary,tempname,&errorcount);
	if(tindex == -1)
	  {
	    fprintf(stderr,"error: %s from %s not in dictionary.\n",
		    tempname,argv[3]);
	    exit(1);
	  }
	sscanf(line+strlen(tempname)+1,"%f\n", &(lastprices[tindex]));
#ifdef DEBUG2
	fprintf(stderr,"tindex: %d %s: %f\n",tindex,tempname,lastprices[tindex]);
	fflush(stderr);
#endif
      }
    fclose(lastweek);

    if(debug)
      { fprintf(stderr,"read last week's list\n"); fflush(stderr); }
  }

if(argc == 5) /* if there is a debug argument, find the card number for it */
  dbg_id = alookup(dictionary,argv[4],&errorcount);
if(dbg_id != -1)
  fprintf(stderr,"debugging card number %d: %s\n",dbg_id,names[dbg_id]);

/* initialize these arrays */
rawtotal = (float *)malloc(numnames*sizeof(float));
squaretotal = (float *)malloc(numnames*sizeof(float));
hits  = (int *)malloc(numnames*sizeof(int));
prices = (pricelist **)malloc(numnames * sizeof(pricelist));
pends  = (pricelist **)malloc(numnames * sizeof(pricelist));
for(i = 0; i < numnames; i++)
  {
    rawtotal[i] = 0.0;
    squaretotal[i] = 0.0;
    hits[i] = 0;
    prices[i] = (pricelist *)NULL;
    pends[i] = (pricelist *)NULL;
  }

if(debug)
  { fprintf(stderr,"initialized arrays\nreading citations..."); fflush(stderr); }

/* read the lines in the datafile.  for each line, if the line matches some 
 * element in the dictionary, then record the price cited on that line.
 */
while(fgets(line,256,datafile) != NULL)
  {
    int len, j, num_words;
    char **words, *test_str; 
    float the_price = 0.0;
    int the_id = -1;
    int best_count = -1;
    int best_length = -1;

    if((++status % 1000) == 0)  /* make sure we're still running */
      {
	fputc('#',stderr);
	fflush(stderr);
      }
    
    cleanup(line); /* remove any uncool characters from the line */
    words = separate_line(line,&num_words);   /* make an array of words */

/* loop over lengths of subsets of words */
    for(len = 1; len <= num_words; len++)
      {
/* loop over all the length len subsets of words[] */
	for(j = 0; j <= num_words-len; j++)
	  {
	    int k, size = 1;
	    for(k = j; k < j+len; k++)  /* loop over a subset starting at j */
	      size += strlen(words[k])+1;
	    test_str = (char *)malloc(size*sizeof(char));
	    *test_str = '\0';  /* force test_str to be a null string */

	    for(k = j; k < j+len; k++)  /* build the appropriate substring */
	      {
		strcat(test_str,words[k]);
		strcat(test_str," ");
	      }
	    size = strlen(test_str);
	    test_str[size-1] = '\0'; /*nuke the space on the end*/
	    size--;  /* adjust size for tests below */
	    switch(test_str[size-1]) /* get rid of trailing punctuation */
	      {
	      case ':': case '.': case ',': case '-':
		test_str[size-1] = '\0';
		size--;
		break;
	      default:
		break;
	      }

/* alookup() test_str, and see if it matches anything: */
	    errorcount = 0; /* reset this */
	    if((id = alookup(dictionary,test_str,&errorcount)) != -1)
	      {
		if (id == -2) /* explicitely ignore this string; */
		  the_id = -1;
		else if (errorcount <= 2) /* throw away really bad matches */
		  {
		    if((best_count == -1) ||
		       ((best_count == errorcount) && (size > best_length)) ||
		       ((errorcount - best_count) <= (size - best_length)/5))
		      {
			the_id = id;
			best_count = errorcount;
			best_length = size;
		      }
		  }
	      }
	    free(test_str);
	  }
      }

    if(the_id != -1 && 
       (the_price = find_price(words,num_words)) > 0.00)
      {
	/* total the prices and the squares of the prices for later */
	rawtotal[the_id] += the_price;
	squaretotal[the_id] += (the_price * the_price);

	hits[the_id]++; /* how many occurances of this card we've seen */

	if(the_id == dbg_id) fprintf(stderr,"%s\n",line);

	/* linked list code for remembering all the prices for later */
	if(prices[the_id] == NULL) /* need to initialize this linked list */
	  {
	    prices[the_id] = (pricelist *)malloc(sizeof(pricelist));
	    prices[the_id]->price = the_price;
	    prices[the_id]->next = (pricelist *)NULL;
	    pends[the_id] = prices[the_id];
	  }
	else  /* add the price at the end of the list */
	  {
	    pends[the_id]->next = (pricelist *)malloc(sizeof(pricelist));
	    pends[the_id] = pends[the_id]->next;
	    pends[the_id]->price = the_price;
	    pends[the_id]->next = (pricelist *)NULL;
	  }
      }
#ifdef DEBUG3
    else
      fprintf(stdout,"no match: %s\n",line);
#endif
    for(i = 0; i < num_words; i++) free(words[i]);
    free(words);
  }

fprintf(stderr,"writing new list\n");fflush(stderr);

if(!shadis)
  {
    printf("card                            price stddev average    high    low change raw N\n");
    printf("----                            ----- ------ ------- ------- ------ ------ -----\n");
  }
else
  {
    printf("card                            price stddev average    high    low raw N change\n");
    printf("----                            ----- ------ ------- ------- ------ ----- ------\n");
  }

for(i = 0; i < numnames; i++)
  {
    int j;
    pricelist *cursor;

    if(hits[i] > 0) /* only print out cards that appeared this week */
      {
	float variance  = 0.0, stddev   = 0.0;  /* lotsa stat variables */
	float filttotal = 0.0, filthits = 0.0;
	float rawavg = rawtotal[i] / (float)hits[i];
	float filtavg = 0.0;
	float high, low, change;

	/* calculate variance and std dev. */
        variance = (squaretotal[i]/((float)hits[i]) - 
		    (rawavg * rawavg));

	stddev = (variance > 0.0) ? sqrt(variance) : 0.0;

	if(i == dbg_id)
	  fprintf(stderr,"variance: %f stddev: %f ",
		  variance,stddev);

        /* calculate mean within stddev of rawtotal/hits, and keep track of
           high/low price within 4 stddev's */
        cursor = prices[i];
	high = low = rawavg; /* initialize high and low prices */
	while(cursor != NULL)
	  {
	    if(abs(cursor->price - rawavg) <= stddev)
	      {
		filttotal += cursor->price;
		filthits++;
	      }
	    if(abs(cursor->price - rawavg) <= 4 * stddev)
	      {
		if(cursor->price > high) high = cursor->price;
		if(cursor->price < low)  low = cursor->price;
	      }
	    cursor = cursor->next;
	  }
	if (filthits > 0.0)
	  filtavg = filttotal/filthits;
	else   /* if all the data gets filtered out, */
	  {
	    filtavg = rawavg;  /* do the reasonable thing, */
	    fprintf(stderr,"warning: filthits = 0 for %s\n",names[i]);
	  }
	if(i == dbg_id)
	  {
	    fprintf(stderr,"filttotal: %f filthits %f\n",filttotal, filthits),
	    cursor = prices[i];
	    while(cursor != NULL)
	      { fprintf(stderr,"%f ",cursor->price);
		cursor = cursor->next; }
	    fprintf(stderr,"\n");
	    fflush(stderr);
	  }

	/* what's the change from last week? */
	if(lastprices[i] == -1 || !have_old_data)
	  change = 0.00;
	else
	  change = filtavg - lastprices[i];

	tfiltavg += filtavg;
	tstddev  += stddev;
	trawavg  += rawavg;
	thigh    += high;
	tlow     += low;
	tchange  += change;
	thits    += hits[i];

	printf("%s,",names[i]);
	if(shadis)
	  printf("  %6.2f, %5.2f,%6.2f,%6.2f,%6.2f,%3d,%5.2f\n",
		 filtavg, stddev, rawavg, high, low, hits[i], change);
	else
	  {
	    if(strlen(names[i]) < 29) /*if there are any spaces left to print*/
	      for(j = 0; j < 29-strlen(names[i]); j++) printf(" ");
	    printf(" %6.2f, %5.2f, %6.2f, %6.2f, %6.2f, %5.2f, %3d\n",
		   filtavg, stddev, rawavg, high, low, change, hits[i]);
	  }
      }
  }
/* print totals */
if(shadis)
  printf("total,  %6.2f,%5.2f,%6.2f,%6.2f,%6.2f,%3d,%5.2f\n",
	 tfiltavg, tstddev, trawavg, thigh, tlow, thits, tchange);
else
  printf(
"total:                       %6.2f, %5.2f, %6.2f, %6.2f, %6.2f, %5.2f, %3d\n",
	 tfiltavg, tstddev, trawavg, thigh, tlow, tchange, thits);

dealloc_dict(dictionary);
exit(0);
}

#if 0
/* this is a wasteful, but fast, hack.  "alpha" characters in this array are
 * everything that's allowed to preceed another alpha character in a word 
 * without necessitating a word break.
 */
static unsigned char is_alpha[256] = 
{
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* nul   -- ^o  */
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* ^p    -- ^_  */
0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, /* space -- /   */
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0     -- ?   */
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* A     -- P   */
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, /* Q     -- `   */
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* a     -- p   */
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, /* q     -- DEL */

0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* extended ascii... */
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};
#endif

/* map unknown char's to spaces, squeeze multiple spaces into a single
 * space, and separate words from non-words.  example: 
 * "L17Jandor's   Ring     $10.00" -> "L17 Jandor's Ring 10.00"
 */
void cleanup(char *line)
{
char lastchar = ' '; /* start this as a space to eliminate leading spaces */
char c, newchar;     /* c = current char, newchar = c after transforming  */
int i=0,j=0; /* counters; i = position being examined, j = pos. to copy onto */
int len = strlen(line); /* to avoid recalculating this all the time */

for(i=0; i < len; i++)
  {
    c = line[i];
    if (my_index(&c) != -1)   /* leave known char's alone */
      newchar = c;
    else                      /* else convert unknown char's to space */
      newchar = ' ';
#if 0
    if (!is_alpha[(unsigned char)lastchar] &&  /* a non-letter,              */
	 is_alpha[(unsigned char)newchar] &&   /* followed by a letter,      */
	 j > 0)                                /* but not at the beginning,  */
      line[j++] = lastchar = ' ';    /* should be separated from it by a ' ' */
#endif
    if (!(lastchar == ' ' && newchar == ' '))  /* squash multiple spaces */
      line[j++] = lastchar = newchar;
  }

line[j]='\0';
}

char **separate_line(char *line, int *num_words)
{
char **words;
int i, j, count = 0;

*num_words = 0;

for(i=0; i < strlen(line); i++)
  if(line[i] == ' ' || line[i] == '\0')
    (*num_words)++;

words = (char **)malloc((*num_words) * sizeof(char *));

for(i = j = 0; i < strlen(line); i++)
  {
    if(line[i] == ' ' || line[i] == '\0')
      {
	words[count] = (char *)malloc(i-j+1 * sizeof(char));
	strncpy(words[count],&(line[j]),i-j);
	words[count][i-j] = '\0';
	j = i+1;
	count++;
      }
  }

return words;
}

float find_price(char **words, int num_words)
{
int i;
int is_clean, dots;
int digits_after_dot;
int nprices = 0;
float hiprice = 0.0;
float midprice = 0.0;
float tprice = 0.0;
char *c;

for(i = 0; i < num_words; i++) /* for all the words */
  {
    is_clean = 1; /* no bad characters so far... */
    dots = 0;     /* how many . seen so far */
    digits_after_dot = 0;
    for(c=words[i]; *c; c++)
      {
	if(*c == '.') dots++;   /* count number of periods in the word */
	if((*c < '0' || *c > '9') && *c != '.')  /* not a digit or dot? */
	  is_clean = 0;        /* then this word isn't suitable for atof() */
	else if (*c != '.' && dots > 0)/* if it is a digit and there's been */
	  digits_after_dot++;          /* a dot, then count this character */
      }
    if(is_clean && (dots == 1) &&
       (digits_after_dot == 2))  /* if this word is a number */
      {
	tprice = atof(words[i]); /* get its value */
        nprices++;
	if(tprice > hiprice)     /* if it's larger than the previous largest */
	  {
	    midprice = hiprice;  /* mid becomes previous largest */
	    hiprice  = tprice;   /* largest become this price */
	  }
	else if(tprice > midprice) /* if it's between midprice and hiprice */
	  midprice = tprice; /* midprice = this price, keep hiprice */
      }
  }

if((nprices == 1) ||  /* one price on a line, return it */
   (nprices == 2))    /* two prices are almost always min bid, current bid */
  return hiprice;
else                /* more than two are almost always min, curr, buyout */
  return midprice; 
}
