#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "my_index.h"
#include "dict.h"

#define MAX_IO_FILES 20
typedef struct {
  char *inames[MAX_IO_FILES];
  char *onames[MAX_IO_FILES];
  char *datafile;
  int nfiles;
} _arg;

void main(int argc, char **argv);
void parse_command_line(int argc, char **argv, _arg *args);
void cleanup(char *line);
char **separate_line(char *line, int *num_words);
int id_line(char *buffer, dictnode *dictionary);

void main(int argc, char **argv)
{
FILE **inputs,
     **outputs,
     *datafile;

int i, j, done;  /* flags and counters */
int *tally;      /* array of tallies */
int id;          /* what group a line identifies as */
long start;      /* file offset into datafile */
char tag;        /* used when reading input file lines */

char phrase[80], buffer[1024]; /* buffers for reading lines from input files */
_arg args;

dictnode *dictionary;

/* parse command line args */
parse_command_line(argc, argv, &args);

/* initializations */
tally   =   (int *)malloc(args.nfiles     * sizeof(int));
inputs  = (FILE **)malloc(args.nfiles     * sizeof(FILE *));
outputs = (FILE **)malloc((args.nfiles+1) * sizeof(FILE *));

for(i=0; i < args.nfiles; i++)
  {
    inputs[i] = fopen(args.inames[i],"r");
    outputs[i] = fopen(args.onames[i],"w");
    tally[i] = 0;
  }
outputs[args.nfiles] = fopen("ambiguous.dat","w");
datafile = fopen(args.datafile,"r");

dictionary = init_dict(ALPHABET_SIZE, my_index);

/* read input files, build master dictionary */
for(i=0; i < args.nfiles; i++)
  {
    done = 0;
    fscanf(inputs[i],"%*d\n");  /* discard the string count from the file */
    while(!done)
      {
	j = 0;
	if((tag = fgetc(inputs[i])) == EOF) /* read the first char */
	  {                                 /* and test for end of file */
	    done = 1;
	    continue;
	  }

	while((phrase[j++] = fgetc(inputs[i])) != '\n'); /* read the line */
	phrase[j-1] = '\0'; /* smash the newline and null terminate */

	if(tag == '-') /* ignore lines with - tags */
	  continue;

	if(insert_phrase(dictionary,phrase,i) == -1)
	  fprintf(stderr,"error inserting %s.  ignoring\n",phrase);
      }
  }

/* now process all the posts in datafile */
done = 0;
start = ftell(datafile); /* starting position in the file */
while(!done)
  {
    if(fgets(buffer, 1024, datafile) == NULL) done = 1;

    if(!strcmp(buffer,".\n")) /* end of post */
      {
	int highest = 0, /* largest tally */
	    hdiff = 0,   /* largest - second largest tally */
	    hindex = 0;  /* 0..nfiles-1 */
	long here,       /* file index to the end of the post */
	     post_size;  /* here - start */
 	char *post;      /* big buffer to hold the whole post */
	int bytes_read = 0, bytes_written = 0;

	for(i=0; i < args.nfiles; i++)
	  if(tally[i] > highest)
	    {
	      hdiff = tally[i] - highest;
	      highest = tally[i];
	      hindex = i;
	    }

	if((highest < 5) ||                /* if the file is ambiguous */
	   ((float)hdiff < .5*(highest)))
	  hindex = args.nfiles;     /* ambiguous file is the last entry */

	here = ftell(datafile);
	post_size = here-start;

	fseek(datafile,start,SEEK_SET);
	post = (char *)malloc(post_size*sizeof(char));
	bytes_read = fread(post, 
			   (size_t)sizeof(char), 
			   (size_t)post_size, 
			   datafile);
	bytes_written = fwrite(post, 
			       (size_t)sizeof(char),
			       (size_t)bytes_read,
			       outputs[hindex]);
	fflush(inputs[hindex]);

	free(post);
	start = here;
	for(i=0; i < args.nfiles; i++) tally[i] = 0;
      }

    id = id_line(buffer, dictionary);
    if(id != -1) tally[id]++;
  }

fclose(datafile);
for(i=0; i < args.nfiles; i++)
  {
    fclose(inputs[i]);
    fclose(outputs[i]);
  }

}


void parse_command_line(int argc, char **argv, _arg *args)
{
int i = 0;
char *c;
char *ext = ".dat";

for(i = 1; i < argc-1; i++)
  {
    args->inames[i-1] = (char *)malloc(strlen(argv[1]) * sizeof(char));
    args->onames[i-1] = (char *)malloc((strlen(argv[1])+strlen(ext)) * sizeof(char));
    strcpy(args->inames[i-1], argv[i]);
    strcpy(args->onames[i-1], argv[i]);
    for(c = args->onames[i-1]; *c; c++)   /* chop input file name down to */
      if(*c == '.') *c = '\0';            /* its basename */
    strcat(args->onames[i-1], ext);       /* add a .dat */
  }
args->datafile = (char *)malloc(strlen(argv[argc-1]) * sizeof(char));
strcpy(args->datafile, argv[argc-1]);
args->nfiles = argc-2; /* not the program name and not the datafile name */
}

/* map unknown char's to spaces, and squeeze multiple spaces into a single
 * space.  example: 
 * "Jandor's   Ring     $10.00" -> "Jandor's Ring 10.00"
 */
void cleanup(char *line)
{
char lastchar = ' '; /* start this as a space to eliminate leading spaces */
char c, newchar;     /* c = current char, newchar = c after transforming  */
int i=0,j=0; /* counters; i = position being examined, j = pos. to copy onto */

for(i=0; i < strlen(line); i++)
  {
    c = line[i];
    if (my_index(&c) != -1) newchar = c;  /* leave known char's alone */

#if 0
    if ((c >= 'a' && c <= 'z') ||   /* what to leave alone */
	(c >= '0' && c <= '9') ||
	(c == '.')) newchar = c;
    else if (c >= 'A' && c <= 'Z') newchar = c + ('a' - 'A');
#endif

    else newchar = ' ';
    if (!(lastchar == ' ' && newchar == ' ')) line[j++] = newchar;
    lastchar = newchar;
  }

line[j]='\0'; 
}

char **separate_line(char *line, int *num_words)
{
char **words;
int i, j, count = 0;

*num_words = 0;

for(i=0; i < strlen(line); i++)
  if(line[i] == ' ' || line[i] == '\0')
    (*num_words)++;

words = (char **)malloc((*num_words) * sizeof(char *));

for(i = j = 0; i < strlen(line); i++)
  {
    if(line[i] == ' ' || line[i] == '\0')
      {
	words[count] = (char *)malloc(i-j+1 * sizeof(char));
	strncpy(words[count],&(line[j]),i-j);
	words[count][i-j] = '\0';
	j = i+1;
	count++;
      }
  }

return words;
}

int id_line(char *buffer, dictnode *dictionary)
{
int len, i, num_words;
char **words, *test_str; 
int the_id = -1;
int best_count = -1;
int best_length = -1;
int errorcount, id;

cleanup(buffer);      /* remove any uncool characters from the line */
words = separate_line(buffer, &num_words); /* make an array of words */

/* loop over lengths of subsets of words */
for(len = 1; len <= num_words; len++)
  {
    /* loop over all the length len subsets of words[] */
    for(i = 0; i <= num_words-len; i++)
      {
	int k, size = 1;
	for(k = i; k < i+len; k++)  /* loop over a subset starting at i */
	  size += strlen(words[k])+1;
	test_str = (char *)malloc(size*sizeof(char));
	*test_str = '\0';  /* force test_str to be a null string */
	
	for(k = i; k < i+len; k++)  /* build the appropriate substring */
	  {
	    strcat(test_str,words[k]);
	    strcat(test_str," ");
	  }
	test_str[strlen(test_str)-1] = '\0'; /*nuke the space on the end*/
	size--;  /* adjust size for tests below */
	
	/* alookup() test_str, and see if it matches anything: */
	errorcount = 0; /* reset this */
	if((id = alookup(dictionary,test_str,&errorcount)) != -1)
	  {
	    if (errorcount <= 2) /* throw away really bad matches */
	      {
		if((best_count == -1) ||
		   ((errorcount - best_count) <= (size - best_length)/5 ) )
		  {
		    the_id = id;
		    best_count = errorcount;
		    best_length = size;
		  }
	      }
	  }
	free(test_str);
      }
  }
for(i = 0; i < num_words; i++) free(words[i]);
free(words);

return(the_id);
}









